##
##  A short introduction to the UCS/R toolbox 
##

## Filename: tutorial.R
## Modified: Thu Jan 22 18:10:08 2004 (evert)   
##   Author: Stefan Evert
##  Purpose: tutorial introduction to the most important features of UCS/R


## This tutorial gives an introduction to the core features of the UCS/R toolbox
## (manipulation of data sets and graphs for the evaluation of association measures)
## from a user perspective.  Several functions that are used internally and that are
## mainly of interest to the programmer are not covered.

## To follow the tutorial, first start an R process in the UCS/R base directory.
## Read the comments (on lines starting with '##') and execute the command lines
## following each comment block in your R process (either using cut & paste, or
## by typing C-c C-n in Emacs/ESS mode). Each command stands on a single, separate
## text line. The command lines must be executed in the order in which they occur
## in the tutorial, and none of them should be left out.

## Have fun!
##    Stefan Evert, August 2003.


## Load UCS/R configuration and library loading mechanism
source("lib/ucs.R")                     

## This prints a list of the available UCS/R libraries
ucs.library()

## Load the required UCS/R libraries
ucs.library("base")
ucs.library("plots")


## Show table of built-in association measures and the corresponding data set variables:
##   key  = string used to identify AM in UCS/R functions
##   long = long descriptive name for AM
##   var  = (name of) variable in data set that holds association scores 
##   rvar = (name of) variable in data set that holds ranking for this AM
## Association scores for all built-in AMs can be computed with UCS/Perl.
AM

## just the list of AM keys
am.keys()

## information can be accessed through functions
am.key2long("log.likelihood.pv")        # description of AM
am.key2var("log.likelihood.pv")         # data set variable for AM scores
am.key2rvar("log.likelihood.pv")        # data set variable for AM ranking
am.var2key("am.odds.ratio")             # extract AM key from variable name
am.var2key("r.odds.ratio")


## A data set need not contain values for all pre-defined AMs (missing AMs
## will simply be ignored, and cannot be plotted).
## Additional AMs defined in a user-provided data set have to be declared
## with the add.am() function, specifying the key (used in variable names)
## and a longer descriptive name.
add.am(key="krenntropy", long="Brigitte Krenn's entropy")

## Have a look at the new entry. A data set annotated with this AM should
## have a variable named <am.krenntropy>.
AM["krenntropy",]

## Read a UCS data set from an uncompressed (.ds) or compressed (.ds.gz) file.
## Here we read in the German adjective+noun data from Evert, Heid, and Lezius (2000).
## The data set has been annotated with association scores for all measures
## supported by the UCS/Perl system.
AN <- read.ds.gz("DataSet/glaw.scores.ds.gz")

## This data set contains 4652 rows (= pair types) and more than 30 columns (= variables).
dim(AN)

## The 30+ variables comprise the core variables id (a numerical unique ID),
## l1 and l2 (the pair type), f, f1, f2, and N (the frequency signature);
## the user-defined variable n.accept (manually annotated true positives);
## as well as more than 20 associations scores. 
colnames(AN)

## A UCS data set file will usually contain only AM scores but no rankings,
## so we have to add them now with the add.ranks() function.  Association
## measures that are not registered in the AM table will be ignored.
## It is also possible to compute rankings only for some of the measures,
## specified with the optional keys parameter (e.g. keys=c("MI", "MI2")).
## Ties in the rankings are broken randomly (default) or according to ID numbers
## (with randomise=FALSE).
AN <- add.ranks(AN)

## Note that we could have combined the last two lines into a single command:
##   AN <- add.ranks( read.ds.gz("DataSet/glaw.scores.ds.gz") )

## Before we can create precision and recall plots, we must mark true positives
## among the candidates.  This is done by adding a Boolean variable named b.TP
## (which assumes the value TRUE for true positives) to the loaded data set.
## Data sets will often not be annotated with an explicit variable marking
## true positives, but rather one or more other variables from which the
## TPs can be inferred (perhaps according to different criteria).  For instance,
## the AN data set contains an integer variable n.accept specifying how many
## annotators accepted each pair type as a collocation.
## We can now define true positives either as the pairs accepted by one
## annotator at least, or only those which were accepted by all annotators.
AN$b.TP <- AN$n.accept > 0

## Now it is easy to compute the number of true positives in the data set
## and the baseline precision (number of TPs / size of the data set)
TPs <- sum(AN$b.TP)
baseline <- round(100 * TPs / nrow(AN), digits=2)
cat(TPs, "true positives,  baseline precision =", baseline, "%\n")

## Combined plot of the precision curves for several AMs. The first argument to
## precision.plot() is a data set, the second a character vector listing the AMs
## to be plotted (identified by their keys).
precision.plot(AN, c("frequency", "t.score", "log.likelihood", "chi.squared", "MI"))

## Use the following (optional) named arguments to fine-tune the display:
##   y.max    maximum precision value (percent) shown on y axis
##   title    main title of the plot
##   show.nbest  vector of n-best list indicated by vertical lines
##   show.baseline=FALSE  disable baseline precision display
precision.plot(AN, c("frequency", "t.score", "log.likelihood", "chi.squared", "MI"), y.max=60, title="Example plot", show.nbest=c(500,1000))

## We can 'zoom in' to an interesting part of the plot by restricting the
## display rectangle with the x.min, x.max, y.min, and y.max parameters.
## All values are given as percentages.
precision.plot(AN, c("frequency", "t.score", "log.likelihood", "chi.squared", "MI"), x.min=10, x.max=40, y.min=20, y.max=50)

## By default, precision values are only shown for n-best lists with n >= 100
## (because they become very unstable for n < 100).  This default can be overridden
## with the parameter n.first.
precision.plot(AN, c("frequency", "t.score", "log.likelihood", "chi.squared", "MI"), x.min=0, x.max=20, y.min=30, y.max=70, n.first=25, show.nbest=c(25,100))

## Black and white versions of the plots can easily be created by setting bw=TRUE.
precision.plot(AN, c("frequency", "t.score", "log.likelihood", "chi.squared", "MI"), y.max=60, bw=TRUE)

## Theoretically, up to 10 different AMs can be combined into a single plot.  This
## is hardly sensible, especially for black and white plots.
precision.plot(AN, c("frequency", "t.score", "log.likelihood", "chi.squared", "MI", "MI2", "MI3", "Dice", "random", "Fisher.pv"), y.max=60)
precision.plot(AN, c("frequency", "t.score", "log.likelihood", "chi.squared", "MI", "MI2", "MI3", "Dice", "random", "Fisher.pv"), y.max=60, bw=TRUE)

## If a <file> argument is given, the plot will be saved to an EPS file rather than
## displayed on screen. By default, a (more or less) square plotting region is used
## and the EPS graphic will look similar to the standard screen display.
precision.plot(AN, c("frequency", "t.score", "log.likelihood", "chi.squared", "MI"), y.max=60, file="Temp/tut1.eps")

## Try viewing <tut1.eps> with gv or ghostview. You will notice that it takes an
## extremely long time to open and display the EPS file. This is due to the fact that
## each precision curve consists of more than 4,500 vertices for each of the
## possible n-best lists. In order to keep file size reasonable and speed up display,
## precision values should only be computed for every 10th n-best list or so, which is
## done by setting n.step=10 (sometimes, an even larger step width may be desirable,
## or a lower value for small data sets).
## Note that the size of <tut1.eps> is reduced from 360 kBytes to about 40 kBytes.
precision.plot(AN, c("frequency", "t.score", "log.likelihood", "chi.squared", "MI"), y.max=60, file="Temp/tut1.eps", n.step=10)

## If required, the scaling aspect between y and x axis can be explicitly set with
## the aspect parameter.  Use aspect=1.0 for 'true' scaling, or values > 1.0 to 'zoom'
## the precision values.  This parameter only has an effect when the plot is saved
## to an EPS file.
precision.plot(AN, c("frequency", "t.score", "log.likelihood", "chi.squared", "MI"), y.max=60, file="Temp/tut2.eps", n.step=10, aspect=1.0)

## For inclusion in articles, EPS files will have to be black and white.  It is often
## desirable to choose a large font (using <cex>) so that the labels are still readable
## when the plot is scaled down for a two-column layout.  You may also want to hide the
## total number of pair types by setting show.npair=NULL.
precision.plot(AN, c("frequency", "t.score", "log.likelihood", "chi.squared", "MI"), y.max=60, file="Temp/tut3.eps", n.step=20, bw=TRUE, cex=1.3, show.npair=NULL)

## The global variable PAR gives some control over the default values for plots.
## It is a list structure with the following named components:
##   col, lty, lwd           =  line styles of graphs (each is a vector of length 10)
##   bw.col, bw.lty, bw.lwd  =  line styles used in b/w mode
##   n.first, n.step         =  default values for n.first and n.step parameters in precision and recall plots
##   tp.first, tp.step       =  default values for tp.first and tp.step parameters in precision-against-recall plots
##   cex                     =  default value for character expansion factors
## Assignment to componentes of the PAR structure makes it easy to change
## default values for multiple plots, e.g.
##   PAR$cex <- 1.3             (use larger font for labels and legend)
##   PAR$n.step <- 10           (reduce number of vertices in precision graphs)
## and to change the default line styles for precision and recall graphs:
##   PAR$col <- c("black", "red", "blue", "#880088", "#00AA00", "FF8800", ...)
## Vectors of length < 10 are automatically replicated, so the command
##   PAR$lwd <- 4
## can be used to select a larger line width for all (colour) graphs.  Reload the
## "plots" library to return to the intial values, or save PAR to another variable
## (e.g. PAR.save <- PAR) while making temporary changes.
## Here, we will have a look at the default line styles for colour graphs and
## then set n.step to 5 to save processing time for significance tests and confidence
## regions.
PAR$col
PAR$lty
PAR$lwd
PAR$n.step <- 5

## A confidence region can be added to one of the graphs in a precision plot
## by setting the confidence level with the conf parameter and selecting the
## desired association measure with the conf.am paramter.  A value of conf=0.05
## produces a 95% confidence region and conf=0.01 a 99% confidence region.
## Setting conf=TRUE selects the default 95% confidence level.  Note that the
## key in conf.am may be abbreviated to a unique prefix (among the displayed AMs).
precision.plot(AN, c("frequency", "t.score", "log.likelihood"), y.max=60, conf=0.05, conf.am="frequency")
precision.plot(AN, c("frequency", "t.score", "log.likelihood"), y.max=60, conf=0.01, conf.am="freq", bw=TRUE)

## Significance differences between the results of two association measures
## can be identified with an exact version of McNemar's test.  The test parameter
## selects the significance level of the test (e.g. test=0.05 for 95% confidence,
## test=0.01 for 99% confidence, or test=TRUE for the default 95% confidence level),
## and test.am1 and test.am2 select the association measures to be compared (which
## may be abbreviated to unique prefixes).
precision.plot(AN, c("t.score", "log.likelihood"), y.max=60, test=0.05, test.am1="log.l", test.am2="t")

## Tests are carried out for every m-th vertex of the graphs (where each vertex corresponds
## to an n-best lists) and significant differences are marked with yellow triangles in the
## plot.  The test.step parameter can be used to adjust the value of m (which defaults to 10).
## Note that distance between triangles also depends on the density of vertices in the
## graphs, which is controlled by the n.step parameter.  Here, we will show more triangles
## by setting test.step=5. 
precision.plot(AN, c("t.score", "log.likelihood"), y.max=60, test=0.05, test.am1="log.l", test.am2="t", test.step=5)

## As the example below shows, McNemar's test may detect significant differences even when
## the absolute difference between two precision values is very small.  Although such cases are
## statistically significant (i.e. they cannot be explained by chance), they are not
## relevant for practical purposes.  For this example, we will consider an absolute
## difference of at least 2 in the percentages relevant.  Setting test.relevant=2, red
## triangles indicate n-best lists where there is significant evidence for such a relevant
## difference between the two association measures.
precision.plot(AN, c("frequency", "log.likelihood"), y.max=60, test=0.05, test.am1="log.l", test.am2="freq")
precision.plot(AN, c("frequency", "log.likelihood"), y.max=60, test=0.05, test.am1="log.l", test.am2="freq", test.relevant=2)

## The usercode parameter allows us to add some fancy graphics to a plot.  In screen mode,
## we could simply add to the plot after calling the precision.plot() function, but this
## is not possible when the plot is written to an EPS file.  Another advantage of the usercode
## callback is that it is invoked before the legend box is drawn.  Here we use the callback
## to show the theoretical upper limit of the precision values in addition to the baseline.
## The usercode function is invoked with two parameters (and should be declared accordingly):
##   region=c(x.min, x.max, y.min, y.max)
##       extent of the plot area (in user coordinates)
##   pr=list(frequency=<pr.data>, t.score=<pr.data>, ...)
##       list of precision and recall values for each of the association measures
##       shown in the plot; the list is indexed by the AM keys and contains the 
##       data structures returned by the precision.recall() function
## (Note that the following block is a multi-line command.)
precision.plot(AN, c("frequency", "t.score", "log.likelihood", "chi.squared", "MI"), y.max=70,
               usercode=function (region, pr) {
                 n.tp <- max(pr[[1]]$tp)
                 size <- max(pr[[1]]$n)
                 baseline <- n.tp / size * 100  # baseline precision = percentage of TPs
                 x <- seq(floor(baseline), 100, by=1)  # x-coordinates for which theoretical max is < 100%
                 y <- baseline / x * 100 # maximum precision = TPs / n = (TPs/size) / (n/size)
                 lines(x, y, lwd=2, lty="dotted")
               })

## The recall.plot() function displays recall curves, which are especially important for
## extraction of collocations from small corpora, where high recall is often a crucial goal.
recall.plot(AN, c("frequency", "t.score", "log.likelihood", "chi.squared", "MI"))

## There a fewer options for recall graphs than for precision graphs.  In particular,
## no significance tests and confidence limits are implemented, and the plot region
## cannot be restricted (i.e. there is no way of zooming in to a part of the plot).
## The following formatting options are supported:
##   title    main title of the plot
##   cex      magnify text and labels (default=1.0)
##   bw       use default settings for black and white graphics
##   n.step   step width for n-best lists
##   file     save plot to EPS file instead of displaying it on screen
## (there is no aspect option; the plot region is always square = aspect of 1.0)
recall.plot(AN, c("frequency", "t.score", "log.likelihood", "chi.squared", "MI"), file="Temp/tut4.eps", n.step=20, bw=TRUE, cex=1.3)


## The most useful means for assessing the practical relevance of differences between AMs
## is a plot of precision against recall; i.e. for each given recall percentage on the x-axis,
## the shortest n-best list achieving at least this recall is selected and the corresponding
## precision is shown on the y-axis.
recall.precision.plot(AN, c("frequency", "t.score", "log.likelihood", "chi.squared", "MI"), y.max=60)

## recall.precision.plot() offers more or less the same options as precision.plot(). However, 
## significance tests and confidence regions are not supported.
## The parameters tp.first and tp.step refer to the number of TPs in each n-best list (which is proportional
## to the x-axis recall value), and they default to 40 and 1, respectively.  In the following example,
## the first precision value shown corrsponds to an absolute recall of approx. 20 TPs, and the recall
## is incremented in steps of 5 TPs for each vertex of the graph.
recall.precision.plot(AN, c("frequency", "t.score", "log.likelihood", "chi.squared", "MI"), y.max=60, tp.first=20, tp.step=5)

## As usual, precision-agains-recall plots can be saved to EPS files using the file paramter, and
## tp.step helps to keep file size moderate.  Try cex=1.3 and bw=TRUE for inclusion in proceedings.
recall.precision.plot(AN, c("frequency", "t.score", "log.likelihood", "chi.squared", "MI"), y.max=60, cex=1.3, title="'nuther example plot", tp.first=20, tp.step=3, file="Temp/tut5.eps", bw=TRUE)

## We can draw vertical lines corresponding to particular recall values (show.recall) as well as
## diagonal lines corresponding to n-best lists (show.nbest).
recall.precision.plot(AN, c("frequency", "t.score", "log.likelihood", "chi.squared", "MI"), y.max=60, show.recall=c(10,20,50))
recall.precision.plot(AN, c("frequency", "t.score", "log.likelihood", "chi.squared", "MI"), y.max=60, show.nbest=c(500,1000))


## Ordinary precision plot display "cumulative" precision for n-best lists of increasing size.
## In order to understand the specific properties of a given association measures, it can be
## more informative to look at "local" precision, which is computed from the number of TPs
## in a window around the n-th pair type in the ranked data set.  Local precision graphs
## are usually very jagged and only few of them should be combined into a single plot.
local.precision.plot(AN, c("frequency", "log.likelihood"), y.max=70)

## Smoothness of the plots is controlled by the window size W, which is the number of pair
## types to the left and right of the n-th pair that are included in the window.  The "total"
## size of the window is thus 2*W + 1.  Add the parameter window=W to change the default
## window size of W=100.
local.precision.plot(AN, c("frequency", "log.likelihood"), y.max=70, window=250)
